﻿#region

using System;

#endregion

namespace SharpObservation
{
    /// <summary>
    /// Base class for implementing the disposable pattern, for types which expose events.
    /// </summary>
    /// <remarks>
    /// Derive from this class, then implement events as follows:
    /// <code>
    /// private readonly static EventKey&lt;PropertyChangingEventHandler&gt; PropertyChangingEventKey = new EventKey&lt;PropertyChangingEventHandler&gt;();
    /// 
    /// public event PropertyChangingEventHandler PropertyChanging
    /// {
    ///     add { AddHandler(PropertyChangingEventKey, value); }
    ///     remove { RemoveHandler(PropertyChangingEventKey, value); }
    /// }
    /// 
    /// protected void RaiseName()
    /// {
    ///     RaiseEvent(PropertyChangingEventKey, raiseEvent =&gt; raiseEvent(this, new PropertyChangingEventArgs("Name")));
    /// }
    /// </code>
    /// A snippet is available (if installed from Snippets.vsi) to create events, just type: "soevt [tab][tab]"
    /// </remarks>
    public abstract class DisposableEventsObject
          : EventsObject
          , IDisposalInfo
          , IDisposedEvent
    {
        #region IDisposedEvent Members

        /// <summary>
        /// Event Key for managing the Disposing event
        /// </summary>
        internal readonly static EventKey<EventHandler> DisposingEventKey = new EventKey<EventHandler>();
        
        /// <summary>
        ///   Provides notification that this instance is being disposed.
        /// </summary>
        public event EventHandler Disposed
        {
            add { AddHandler(DisposingEventKey, value); }
            remove { RemoveHandler(DisposingEventKey, value); }
        }

        #endregion

        /// <summary>
        ///   True if the object has been disposed.
        /// </summary>
        private bool _isDisposed;

        #region IDisposable Members

        /// <summary>
        ///   Disposes the factory instance
        /// </summary>
        public void Dispose()
        {
            if (_isDisposed) return;
            _isDisposed = true;
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        #endregion

        #region IDisposalInfo Members

        /// <summary>
        ///   Gets whether the factory instance has been disposed.
        /// </summary>
        public bool IsDisposed
        {
            get { return _isDisposed; }
        }

        #endregion

        /// <summary>
        ///  Checks whether the instance has been disposed, and throws an ObjectDisposedException if it has.
        /// </summary>
        /// <exception cref = "ObjectDisposedException">Thrown if instance has been disposed</exception>
        public void ThrowIfDisposed()
        {
            if (_isDisposed)
                throw new ObjectDisposedException(GetType().Name);
        }

        /// <summary>
        /// Finalizes the instance. 
        /// </summary>
        ~DisposableEventsObject()
        {
            if (_isDisposed) return;
            _isDisposed = true;
            Dispose(false);
        }

        /// <summary>
        /// Handles disposal
        /// </summary>
        /// <param name = "disposing">True if Dispose was called. False if finalizing</param>
        protected virtual void Dispose(bool disposing)
        {
            // Raise the Disposing event
            RaiseEvent(DisposingEventKey, raiseEvent => raiseEvent(this, EventArgs.Empty));
        }
  
    }
}