﻿#region

using System;

#endregion

namespace SharpObservation
{
    /// <summary>
    /// Base class for implementing the an object which exposes events and implements the disposable pattern
    /// </summary>
    /// <remarks>
    /// Derive from this class, to implement the disposable pattern correctly.
    /// </remarks>
    public abstract class DisposableObject : IDisposalInfo
    {
        /// <summary>
        ///   Gets whether the factory instance has been disposed.
        /// </summary>
        public bool IsDisposed { get; private set; }
        
        /// <summary>
        ///   Disposes the factory instance
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1063:ImplementIDisposableCorrectly")]
        public void Dispose()
        {
            if (!IsDisposed)
            {
                IsDisposed = true;
                Dispose(true);
                GC.SuppressFinalize(this);
            }
        }

        /// <summary>
        ///   Finalizes the factory instance. Should never be called.
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Microsoft.Design", "CA1063:ImplementIDisposableCorrectly")]
        ~DisposableObject()
        {
            if (!IsDisposed)
            {
                IsDisposed = true;
                Dispose(false);
            }
        }


        /// <summary>
        ///   Checks whether the factory instance has been disposed, and throws an ObjectDisposedException if it has
        /// </summary>
        /// <exception cref = "ObjectDisposedException">Thrown if instance has been disposed</exception>
        public void CheckIsDisposed()
        {
            if (IsDisposed)
                throw new ObjectDisposedException(GetType().Name);
        }


        /// <summary>
        /// Releases resources
        /// </summary>
        /// <param name="disposing"><c>true</c> to release both managed and unmanaged resources; <c>false</c> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            // this method could be abstract, but is deliberately implemented to support refactoring...
            // derived classes can always call base.Dispose.
        }
    }
}