﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SharpObservation
{
    /// <summary>
    /// Base type for implementing observable objects which are disposable
    /// </summary>
    /// <remarks>
    /// <para>The following example shows how to create an observable class with a name property:</para>
    /// <code>
    /// public class SimpleButObservable : DisposableObservableObject
    /// {
    ///      private string _name;
    ///     
    ///      private static readonly Property&lt;SimpleButObservable, string&gt; NameProperty =
    ///                new Property&lt;SimpleButObservable, string&gt;(
    ///                    c => c.Name,
    ///                    c => c._name,
    ///                    (c, v) => c._name = v);
    ///     
    ///      public string Name
    ///      {
    ///          get { return _name; }
    ///          set { NameProperty.Set(this, value); }
    ///      }
    /// }
    /// </code>
    /// </remarks>
    public abstract class DisposableObservableObject : ObservableObject
          , IDisposalInfo
          , IDisposedEvent
    {
        #region IDisposedEvent Members

        /// <summary>
        /// Event Key for managing the Disposing event
        /// </summary>
        internal static readonly EventKey<EventHandler> DisposingEventKey = DisposableEventsObject.DisposingEventKey;

        /// <summary>
        /// Provides notification that this instance is being disposed.
        /// </summary>
        public event EventHandler Disposed
        {
            add { AddHandler(DisposingEventKey, value); }
            remove { RemoveHandler(DisposingEventKey, value); }
        }

        #endregion

        /// <summary>
        /// True if the object has been disposed.
        /// </summary>
        private bool _isDisposed;


        /// <summary>
        ///  Checks whether the instance has been disposed, and throws an ObjectDisposedException if it has.
        /// </summary>
        /// <exception cref = "ObjectDisposedException">Thrown if instance has been disposed</exception>
        public void ThrowIfDisposed()
        {
            if (_isDisposed)
                throw new ObjectDisposedException(GetType().Name);
        }


        #region IDisposalInfo Members

        /// <summary>
        ///   Gets whether the factory instance has been disposed.
        /// </summary>
        public bool IsDisposed
        {
            get { return _isDisposed; }
        }

        #endregion


        #region IDisposable Members

        /// <summary>
        ///   Disposes the factory instance
        /// </summary>
        public void Dispose()
        {
            if (_isDisposed) return;
            // Raise the Disposing event
            RaiseEvent(DisposingEventKey, raiseEvent => raiseEvent(this, EventArgs.Empty));

            _isDisposed = true;
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Finalizes the instance. 
        /// </summary>
        ~DisposableObservableObject()
        {
            if (_isDisposed) return;
            // Raise the Disposing event
            RaiseEvent(DisposingEventKey, raiseEvent => raiseEvent(this, EventArgs.Empty));

            _isDisposed = true;
            Dispose(false);
        }


        /// <summary>
        /// Handles disposal
        /// </summary>
        /// <param name = "disposing">True if Dispose was called. False if finalizing</param>
        protected virtual void Dispose(bool disposing)
        {
        }

        #endregion

    }
}
