﻿using System;
using System.Linq.Expressions;
using System.Reflection;

namespace SharpObservation
{
    static class ExpressionExtenstions
    {
        private class AssignmentHelper<T>
        {
// ReSharper disable UnusedMember.Local
            private static void SetValue(ref T target, T value)
// ReSharper restore UnusedMember.Local
            {
                target = value;
            }

            internal static readonly MethodInfo MethodInfoSetValue =
                typeof(AssignmentHelper<T>).GetMethod("SetValue", BindingFlags.NonPublic | BindingFlags.Static);
        }


        /// <summary>
        /// Converts a get expression into an assignment expression
        /// </summary>
        /// <typeparam name="TInstance">The interfaceType of the object on which the field is declared</typeparam>
        /// <typeparam name="TProp">The interfaceType of the field to set</typeparam>
        /// <param name="fieldGetter"></param>
        /// <returns>
        /// An expression that assigns the field that was returned by the input expression
        /// </returns>
        /// <example>
        ///     var assignExpression = ToFieldAssignExpression( Customer c => c.Age );
        ///     var assignName = assignExpression.Compile();
        ///     assignName("John Smith");
        /// </example>
        public static Expression<Action<TInstance, TProp>> ToFieldAssignExpression<TInstance, TProp>
                (this Expression<Func<TInstance, TProp>> fieldGetter)
        {
            if (fieldGetter == null)
                throw new ArgumentNullException(@"fieldGetter");

            if(fieldGetter.Parameters.Count != 1 || !(fieldGetter.Body is MemberExpression))
                throw new ArgumentException ("Input expression must be a single parameter field getter, e.g. g => g._fieldToSet  or function(g) g._fieldToSet");

            var parms = new ParameterExpression[] {
                                        fieldGetter.Parameters[0], 
                                        Expression.Parameter(typeof(TProp), "value")};

            Expression body = Expression.Call(AssignmentHelper<TProp>.MethodInfoSetValue,
                                              new Expression[] { fieldGetter.Body, parms[1] });

            return Expression.Lambda<Action<TInstance, TProp>>(body, parms);
        }


        /// <summary>
        /// Converts a get expression into an assignment delegate
        /// </summary>
        /// <typeparam name="TInstance">The interfaceType of the object on which the field is declared</typeparam>
        /// <typeparam name="TProp">The interfaceType of the field to set</typeparam>
        /// <param name="fieldGetter"></param>
        /// <returns>
        /// A delegate that assigns the field that was returned by the input expression
        /// </returns>
        /// <example>
        ///     var assignName = ToFieldAssignment( Customer c => c.Age );
        ///     assignName("John Smith");
        /// </example>
        public static Action<TInstance, TProp> ToFieldAssignment<TInstance, TProp>
            (this Expression<Func<TInstance, TProp>> fieldGetter)
        {
            return fieldGetter.ToFieldAssignExpression().Compile();
        }
    }
}
