#region

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Diagnostics.Contracts;
using System.Linq.Expressions;
using System.Threading;

#endregion

namespace SharpObservation
{
    /// <summary>
    /// Base for implementations of the INotifyPropertyChanging and INotifyPropertyChanged pattern.
    /// </summary>
    /// <remarks>
    /// <para>
    ///   ObservableObject serves as a base class for derived types that implement the property change notiofication pattern.
    ///   In addition to the standard .NET property change notifications, this class provides several useful features:</para>
    /// <list type="bullet">
    /// <item>Centralized property change implementation which bypassing event firings if no listeners are attached.</item>
    /// <item>Ability to undo property changes.</item>
    /// <item>IsNotificationActive property indicates whether the object is being listened to.</item>
    /// </list> 
    /// <para>
    /// A snippet is available (if installed from Snippets.vsi) to create properties, just type: "soprop [tab][tab]".
    /// </para>
    /// <para>The following example shows how to create an observable class with a name property:</para>
    /// <code>
    /// public class SimpleButObservable : ObservableObject
    /// {
    ///      private string _name;
    ///     
    ///      private static readonly Property&lt;SimpleButObservable, string&gt; NameProperty =
    ///                new Property&lt;SimpleButObservable, string&gt;(
    ///                    c => c.Name,
    ///                    c => c._name,
    ///                    (c, v) => c._name = v);
    ///     
    ///      public string Name
    ///      {
    ///          get { return _name; }
    ///          set { NameProperty.Set(this, value); }
    ///      }
    /// }
    /// </code>
    /// </remarks>
    public abstract class ObservableObject
        : EventsObject
          , INotifyPropertyChanging
          , INotifyPropertyChanged
    {
        #region ChangeListenerCount

        /// <summary>
        ///   Tracks the number of combined changing/changed delegates that are attached
        /// </summary>
        private int _changeListenerCount;

        /// <summary>
        ///   Tracks the number of combined changing/changed delegates that are attached
        /// </summary>
        protected int ChangeListenerCount
        {
            get { return _changeListenerCount; }
        }

        #endregion

        #region Event Key Registrations (Static)

        /// <summary>
        ///   Event Key for PropertyChanging Event
        /// </summary>
        protected static readonly EventKey<PropertyChangingEventHandler> PropertyChangingEventKey =
            new EventKey<PropertyChangingEventHandler>();

        /// <summary>
        ///   Event Key for PropertyChanged Event
        /// </summary>
        protected static readonly EventKey<PropertyChangedEventHandler> PropertyChangedEventKey =
            new EventKey<PropertyChangedEventHandler>();

        /// <summary>
        ///   Event Key for IsNotificationActiveChanged Event
        /// </summary>
        protected static readonly EventKey<IsNotificationActiveChangedEventHandler> IsNotificationActiveChangedEventKey
            = new EventKey<IsNotificationActiveChangedEventHandler>();

        #endregion

        #region CLR Event declarations

        #region INotifyPropertyChanged Members

        /// <summary>
        ///   Notifies listeners that a property has been changed.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged
        {
            // Add the input delegate to the collection.
            add
            {
                AddHandler(PropertyChangedEventKey, value);
                ListenerAdded();
            }
            // Remove the input delegate from the collection.
            remove
            {
                RemoveHandler(PropertyChangedEventKey, value);
                ListenerRemoved();
            }
        }

        #endregion

        #region INotifyPropertyChanging Members

        /// <summary>
        ///   Notifies listeners that a property is about to change.
        /// </summary>
        public event PropertyChangingEventHandler PropertyChanging
        {
            // Add the input delegate to the collection.
            add
            {
                AddHandler(PropertyChangingEventKey, value);
                ListenerAdded();
            }
            // Remove the input delegate from the collection.
            remove
            {
                RemoveHandler(PropertyChangingEventKey, value);
                ListenerRemoved();
            }
        }

        #endregion

        /// <summary>
        /// Fired whenever <c ref = "isNotificationActive" /> changes.
        /// </summary>
        public event IsNotificationActiveChangedEventHandler IsNotificationActiveChanged
        {
            add { AddHandler(IsNotificationActiveChangedEventKey, value); }
            remove { RemoveHandler(IsNotificationActiveChangedEventKey, value); }
        }

        #endregion

        #region Event Raising Methods

        #region OnPropertyChanging

        /// <summary>
        ///   Sends an "all property" changing notification to all listeners
        /// </summary>
        protected void OnPropertyChanging()
        {
            if (ChangeListenerCount > 0)
                OnPropertyChanging(new PropertyChangingEventArgs(null));
        }

        /// <summary>
        ///   Sends a property changing notification to all listeners
        /// </summary>
        protected virtual void OnPropertyChanging(PropertyChangingEventArgs args)
        {
            Contract.Requires(args.PropertyName == null || GetType().GetProperty(args.PropertyName) != null);

            if (ChangeListenerCount > 0)
                RaiseEvent(PropertyChangingEventKey, raiseEvent => raiseEvent(this, args));
        }

        /// <summary>
        ///   Sends a property changing notification to all listeners
        /// </summary>
        /// <param name = "propertyName"></param>
        protected void OnPropertyChanging(string propertyName)
        {
            Contract.Requires(propertyName == null || GetType().GetProperty(propertyName) != null);

            if (ChangeListenerCount > 0)
                OnPropertyChanging(new PropertyChangingEventArgs(propertyName));
        }

        /// <summary>
        /// Sends a property changed notification to all listeners
        /// </summary>
        /// <param name = "property">The property that was changed</param>
        protected void OnPropertyChanging<TInstance, TProp>(Property<TInstance, TProp> property)
            where TInstance : ObservableObject
        {
            OnPropertyChanging(new PropertyChangingEventArgs(property.Name));
        }

        /// <summary>
        ///   Sends a property changing notification to all listeners, providing the richly typed PropertyChangingEventArgs
        /// </summary>
        /// <typeparam name = "T">Factory type</typeparam>
        /// <param name = "propertyName">The name of the property that was changing</param>
        /// <param name = "oldValue">The factory Value of the property</param>
        /// <param name = "newValue">The new Value of the property</param>
        protected void OnPropertyChanging<T>(string propertyName, T oldValue, T newValue)
        {
            if (ChangeListenerCount > 0)
                OnPropertyChanging(new PropertyChangingEventArgs<T>(propertyName, oldValue, newValue));
        }

        #endregion

        #region OnPropertyChanged

        /// <summary>
        ///   Sends an "all property" changed notification to all listeners
        /// </summary>
        protected void OnPropertyChanged()
        {
            if (ChangeListenerCount > 0)
                OnPropertyChanged(new PropertyChangedEventArgs(null));
        }

        /// <summary>
        ///   Sends a property changed notification to all listeners
        /// </summary>
        protected virtual void OnPropertyChanged(PropertyChangedEventArgs args)
        {
            Contract.Requires(args.PropertyName == null || GetType().GetProperty(args.PropertyName) != null);

            if (ChangeListenerCount > 0)
                RaiseEvent(PropertyChangedEventKey, propertyChanged => propertyChanged(this, args));
        }

        /// <summary>
        ///   Sends a property changed notification to all listeners
        /// </summary>
        /// <param name = "propertyName">The name of the property that was changed</param>
        protected void OnPropertyChanged(string propertyName)
        {
            OnPropertyChanged(new PropertyChangedEventArgs(propertyName));
        }
           
        /// <summary>
        ///   Sends a property changed notification to all listeners
        /// </summary>
        /// <param name = "property">The property that was changed</param>
        protected void OnPropertyChanged<TInstance, TProp>(Property<TInstance, TProp> property)
            where TInstance : ObservableObject
        {
            OnPropertyChanged(new PropertyChangedEventArgs(property.Name));
        }
        
        /// <summary>
        ///   Sends a property changed notification to all listeners, providing the richly typed PropertyChangedEventArgs
        /// </summary>
        /// <typeparam name = "T">Factory type</typeparam>
        /// <param name = "propertyName">The name of the property that was changed</param>
        /// <param name = "oldValue">The factory Value of the property</param>
        /// <param name = "newValue">The new Value of the property</param>
        protected void OnPropertyChanged<T>(string propertyName, T oldValue, T newValue)
        {
            if (ChangeListenerCount > 0)
                OnPropertyChanged(new PropertyChangedEventArgs<T>(propertyName, oldValue, newValue));
        }

        #endregion

        #region OnIsNotificationActiveChanged

        /// <summary>
        /// Raises the IsNotificationActiveChanged event
        /// </summary>
        protected virtual void OnIsNotificationActiveChanged()
        {
            RaiseEvent(
                IsNotificationActiveChangedEventKey,
                raiseEvent => raiseEvent(this, new IsNotificationActiveChangedEventArgs(ChangeListenerCount > 0))
                );
        }

        #endregion

        #endregion

        #region Notification Activation/Deactivation - IsNotificationActive

        /// <summary>
        ///   Gets whether any delegates are attached to PropertyChanging or PropertyChanged
        /// </summary>
        public bool IsNotificationActive
        {
            get { return ChangeListenerCount > 0; }
        }


        /// <summary>
        ///   Invoked when a delegate is attached to PropertyChanged or PropertyChanging
        /// </summary>
        private void ListenerAdded()
        {
            if ((_changeListenerCount = ChangeListenerCount + 1) == 0)
                OnIsNotificationActiveChanged();
        }

        /// <summary>
        ///   Invoked when a delegate is dettached to PropertyChanged or PropertyChanging
        /// </summary>
        private void ListenerRemoved()
        {
            if ((_changeListenerCount = ChangeListenerCount - 1) == 0)
                OnIsNotificationActiveChanged();
        }

        #endregion

        #region DoSetProperty

        /// <summary>
        /// Executes the changing of a property from one Value to another.
        /// </summary>
        /// <typeparam name = "TInstance">The type of the owning class</typeparam>
        /// <typeparam name = "TProp">The type of the property being set</typeparam>
        /// <param name = "propertyName">The name of the property which is being changed</param>
        /// <param name = "oldValue">The old value of the property</param>
        /// <param name = "newValue">The (desireD) new value of the property</param>
        /// <param name = "setter">The delegate which sets the backing field</param>
        protected internal virtual void DoSetProperty<TInstance, TProp>(
            string propertyName,
            TProp oldValue,
            TProp newValue,
            Action<TInstance, TProp> setter)
            where TInstance : ObservableObject
        {
            Contract.Requires(propertyName != null);

            var thisAsTInstance = this as TInstance;

            Debug.Assert(thisAsTInstance != null,
                         @"DoSetProperty can only be invoked for properties declared on this instance's type - in this case " +
                         GetType().Name);

            if (ChangeListenerCount > 0)
            {
                OnPropertyChanging(new PropertyChangingEventArgs<TProp>(propertyName, oldValue, newValue));

                setter(thisAsTInstance, newValue);

                OnPropertyChanged(new UndoablePropertyChangedEventArgs<TInstance, TProp>(propertyName, oldValue, newValue, thisAsTInstance, setter));
            }
            else
            {
                setter(thisAsTInstance, newValue);
            }
        }


        /// <summary>
        /// Sets the value of a property.
        /// </summary>
        /// <typeparam name="TInstance">The type of the instance owning the property</typeparam>
        /// <typeparam name="TProp">The type of the prop.</typeparam>
        /// <param name="property">The property.</param>
        /// <param name="newValue">The new value.</param>
        protected void SetProperty<TInstance, TProp>(Property<TInstance, TProp> property, TProp newValue)
            where TInstance : ObservableObject
        {
            property.Set((TInstance)this, newValue);
        }

        #endregion


        #region Property Declaration & Supporting Types

        /// <summary>
        /// Encapsulates details about a property that raises change notifications.
        /// </summary>
        /// <typeparam name="TInstance">The type on which the property is declared</typeparam>
        /// <typeparam name="TProp">The type of the property value</typeparam>
        public class Property<TInstance, TProp>
            where TInstance : ObservableObject
        {
            /// <summary>
            /// The equality test delegate
            /// </summary>
            internal readonly Func<TProp, TProp, bool> _equals;

            /// <summary>
            /// The getter delegate
            /// </summary>
            internal readonly Func<TInstance, TProp> _getter;

            /// <summary>
            /// The name of the property
            /// </summary>
            internal readonly string Name;

            /// <summary>
            /// The setter delegate
            /// </summary>
            internal readonly Action<TInstance, TProp> _setter;

            /// <summary>
            /// Creates a new property declaration
            /// </summary>
            /// <param name="fnGetProperty">Expression which returns the property member</param>
            /// <param name="fnGetField">Expression which returns the field member</param>
            public Property(Expression<Func<TInstance, TProp>> fnGetProperty,
                        Expression<Func<TInstance, TProp>> fnGetField)
            {
                var visitor = new PropertyPathExtractor();
                visitor.Visit(fnGetProperty);
                Debug.Assert(visitor.CurrentExpression != null);
                Name = visitor.CurrentExpression.Member.Name;
                _getter = fnGetField.Compile();
                _setter = fnGetField.ToFieldAssignment();
                _equals = EqualityComparer<TProp>.Default.Equals;
            }

            /// <summary>
            /// Creates a new property declaration
            /// </summary>
            /// <param name="name">The name to report during property change notifications</param>
            /// <param name="getter">delegate which returns the field value</param>
            /// <param name="setter">delegate which sets the field value</param>
            public Property(string name, Func<TInstance, TProp> getter, Action<TInstance, TProp> setter)
            {
                Name = name;
                _getter = getter;
                _setter = setter;
                _equals = EqualityComparer<TProp>.Default.Equals;
            }

            /// <summary>
            /// Creates a new property declaration
            /// </summary>
            /// <param name="fnGetProperty">Expression which returns the property member</param>
            /// <param name="getter">delegate which returns the field value</param>
            /// <param name="setter">delegate which sets the field value</param>
            public Property(Expression<Func<TInstance, TProp>> fnGetProperty,
                        Func<TInstance, TProp> getter,
                        Action<TInstance, TProp> setter)
            {
                var visitor = new PropertyPathExtractor();
                visitor.Visit(fnGetProperty);
                Debug.Assert(visitor.CurrentExpression != null);
                Name = visitor.CurrentExpression.Member.Name;
                _getter = getter;
                _setter = setter;
                _equals = EqualityComparer<TProp>.Default.Equals;
            }


            /// <summary>
            /// Sets the underlying field and raises proeprty change notifications
            /// </summary>
            /// <param name="instance">The instance owning the property being changed</param>
            /// <param name="value">The new value of the property</param>
            public void Set(TInstance instance, TProp value)
            {
                TProp currentValue = _getter(instance);

                if (!_equals(currentValue, value))
                    instance.DoSetProperty(Name, currentValue, value, _setter);
            }
        }

        #endregion
    }
}