﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

#endregion

namespace SharpObservation
{
    /// <summary>
    ///   A collection of of items which are indirectly referenceds
    /// </summary>
    /// <typeparam name = "TItem">The type of item in the collection</typeparam>
    /// <typeparam name = "TReference">The type of weak reference to use</typeparam>
    /// <remarks>
    ///   ReferenceCollection is a collection of items where the job of referencing the items is defered to 
    ///   another type. This allows, for example, the WeakReference type to be used to reference the items
    ///   in the collection, effectively making the entire collection weak referenced.
    /// </remarks>
    public class ReferenceCollection<TItem, TReference> : ICollection<TItem>
        where TItem : class
        where TReference : IReference<TItem>
    {
        #region Backing Fields

        private readonly List<TReference> _collection;
        private readonly Func<TItem, TReference> _factory;
        private int _addInvokeCounter;

        #endregion

        #region Constructor

        /// <summary>
        ///   Initializes a new instance of the <see cref = "ReferenceCollection{TItem,TReference}" /> class.
        /// </summary>
        /// <param name = "factory">The factory.</param>
        public ReferenceCollection(Func<TItem, TReference> factory)
        {
            _factory = factory;
            _collection = new List<TReference>();
        }

        /// <summary>
        ///   Initializes a new instance of the <see cref = "ReferenceCollection{TItem,TReference}" /> class.
        /// </summary>
        /// <param name = "factory">The factory.</param>
        /// <param name = "collection">An existing collection to initialize from</param>
        public ReferenceCollection(Func<TItem, TReference> factory, IEnumerable<TItem> collection)
        {
            _factory = factory;

            _collection = collection == null
                              ? new List<TReference>()
                              : new List<TReference>(collection.Select(factory));
        }

        #endregion

        #region ICollection<TItem> Members

        /// <summary>
        ///   Adds the specified item.
        /// </summary>
        /// <param name = "item">The item.</param>
        public void Add(TItem item)
        {
            var weakRef = _factory(item);

            // every 1000 adds, do a cleanup check
            if (++_addInvokeCounter > 1000)
            {
                _addInvokeCounter = 0;
                for (int i = 0; i < _collection.Count; i++)
                {
                    var ref2 = _collection[i];
                    if (!ref2.IsAlive)
                        _collection.Remove(ref2);
                }
            }
            _collection.Add(weakRef);
        }

        /// <summary>
        ///   Clears this instance.
        /// </summary>
        public void Clear()
        {
            _collection.Clear();
            _addInvokeCounter = 0;
        }

        /// <summary>
        ///   Determines whether [contains] [the specified item].
        /// </summary>
        /// <param name = "item">The item.</param>
        /// <returns>
        ///   <c>true</c> if [contains] [the specified item]; otherwise, <c>false</c>.
        /// </returns>
        public bool Contains(TItem item)
        {
            var contains = _collection.Any(t => t.IsAlive && t.Target == item);
            return contains;
        }

        /// <summary>
        ///   Copies to.
        /// </summary>
        /// <param name = "array">The array.</param>
        /// <param name = "arrayIndex">Index of the array.</param>
        public void CopyTo(TItem[] array, int arrayIndex)
        {
            _collection.Where(t => t.IsAlive).ToArray().CopyTo(array, arrayIndex);
        }

        /// <summary>
        ///   Gets the count.
        /// </summary>
        /// <value>The count.</value>
        public int Count
        {
            get { return _collection.Count; }
        }

        /// <summary>
        ///   Gets a value indicating whether this instance is read only.
        /// </summary>
        /// <value>
        ///   <c>true</c> if this instance is read only; otherwise, <c>false</c>.
        /// </value>
        public bool IsReadOnly
        {
            get { return false; }
        }

        /// <summary>
        ///   Removes the specified item.
        /// </summary>
        /// <param name = "item">The item.</param>
        /// <returns></returns>
        public bool Remove(TItem item)
        {
            bool removed = false;
            foreach (var i in _collection.ToArray())
            {
                if (!i.IsAlive)
                    _collection.Remove(i);
                else if (i.Target == item)
                {
                    removed = _collection.Remove(i);
                }
            }
            return removed;
        }

        /// <summary>
        ///   Gets the enumerator.
        /// </summary>
        /// <returns></returns>
        public IEnumerator<TItem> GetEnumerator()
        {
            return _collection.Where(i => i.IsAlive).Select(i => i.Target).GetEnumerator();
        }

        /// <summary>
        ///   Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        ///   An <see cref = "T:System.Collections.IEnumerator" /> object that can be used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        #endregion
    }
}