﻿#region

using System;
using System.ComponentModel;
using System.Diagnostics.Contracts;
using System.Linq.Expressions;
using System.Reflection;
using System.Threading;

#endregion

namespace SharpObservation
{
    //TODO: Rewrite to use similar MSIL semantics as WeakDelegate uses. For more flxeible.

    ///<summary>
    ///  Extends EventHandler with methods that convert the delegates to hold weak references to their target instances.
    ///</summary>
    ///<remarks>
    ///  A common problem with the standard CLR event model is that events cause a reverse reference to exist whereby the "observed object" 
    ///  (i.e. the instance raising the event) implicitly references the observing object (the instance that contains the method invoked by the delegate).
    ///  DelegateExtensions extends EventHandler, PropertyChangedEventHandler and PropertyChangingEventHandler with the AsWeak and AsWeakAsync methods.
    ///  These methods provide a resonably performant solution to the aforementioned reverse reference anti-pattern, by converting a delegate into a 
    ///  "weak reference" delegate which does not keep the observer alive.
    ///
    ///  AsWeak creates a weak reference delegate which (optionally) attempts to invoke the tagrte method using a SynchronizationContext.
    ///  AsWeakAsync creates a weak reference delegate which invokes the target method asynchronously of the thread which raised the event - either by using 
    ///  ThreadPool.QueueWorkUserItem, or by using SynchronizationContext.Post if a synchronization context is available.
    /// 
    ///  The underlying implementation of constructs instances of WeakEventHandler or ContextWeakEventHandler in order to correctly manage the weak reference
    ///  and resolve synchronization.
    /// 
    ///  While every effort has been made to make this implementation performant (and it is indeed far more performant than the .NET WeakEventManager pattern),
    ///  it can still be [marginally] out-performed by a well-written Dispose pattern on the part of the observer.
    /// 
    ///  A more performant solution to the weak referencing event pattern is for a class to simply 
    ///  implement a proxy class (derived from WeakReference) which exposes the required callback methods, 
    ///  and have those methods in tern cass control on to the target instance.
    ///</remarks>
    public static partial class SynchronizedDelegates
    {
        #region EventHandler<>.AsAsync

        /// <summary>
        ///   Wraps the eventHandler so that it is invoked asynchronously, and on the caller's thread if a SynchronizationContext is available.
        /// </summary>
        /// <typeparam name = "TEvent">The typeToMatch of event arg for the EventHandler</typeparam>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <returns>A delegate wrapper which invokes inner delegate asynchronously</returns>
        public static EventHandler<TEvent> AsAsync<TEvent>(this EventHandler<TEvent> eventHandler)
            where TEvent : EventArgs
        {
            return AsAsync(eventHandler, SynchronizationContext.Current);
        }

        /// <summary>
        ///   Wraps the eventHandler so that it is invoked asynchronously, and on the caller's thread if a SynchronizationContext is available.
        /// </summary>
        /// <typeparam name = "TEvent">The typeToMatch of event arg for the EventHandler</typeparam>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <param name = "syncContext">Optional. The context in which the callback is posted</param>
        /// <returns>A delegate wrapper which invokes inner delegate asynchronously</returns>
        public static EventHandler<TEvent> AsAsync<TEvent>(this EventHandler<TEvent> eventHandler,
                                                           SynchronizationContext syncContext)
            where TEvent : EventArgs
        {
            if (syncContext != null)
            {
                return (o, e) =>
                           {
                               var sopcb = new SendOrPostCallback(b => eventHandler(b, e));
                               syncContext.Post(sopcb, o);
                           };
            }
            return (o, e) =>
                       {
                           var sopcb = new WaitCallback(b => eventHandler(b, e));
                           ThreadPool.QueueUserWorkItem(sopcb, o);
                       };
        }

        #endregion

        #region EventHandler<>.AsSend

        /// <summary>
        ///   Wraps the eventHandler so that it is invoked synchronously on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <typeparam name = "TEvent">The typeToMatch of event arg for the EventHandler</typeparam>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <returns>A delegate wrapper which invokes inner delegate synchronously</returns>
        public static EventHandler<TEvent> AsSend<TEvent>(this EventHandler<TEvent> eventHandler)
            where TEvent : EventArgs
        {
            return AsSend(eventHandler, SynchronizationContext.Current);
        }

        /// <summary>
        ///   Wraps the eventHandler so that it is invoked synchronously on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <typeparam name = "TEvent">The typeToMatch of event arg for the EventHandler</typeparam>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <param name = "syncContext">Optional. The context in which the callback is sent</param>
        /// <returns>A delegate wrapper which invokes inner delegate synchronously</returns>
        public static EventHandler<TEvent> AsSend<TEvent>(this EventHandler<TEvent> eventHandler,
                                                          SynchronizationContext syncContext)
            where TEvent : EventArgs
        {
            if (syncContext != null)
            {
                return (o, e) =>
                           {
                               var sopcb = new SendOrPostCallback(b => eventHandler(b, e));
                               syncContext.Send(sopcb, o);
                           };
            }
            return eventHandler;
        }

        #endregion
        
        #region PropertyChangingEventHandler.AsAsync

        /// <summary>
        ///   Wraps the PropertyChangingEventHandler so that it is invoked asynchronously, and on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <returns>A delegate wrapper which invokes inner delegate asynchronously</returns>
        public static PropertyChangingEventHandler AsAsync(this PropertyChangingEventHandler eventHandler)
        {
            return AsAsync(eventHandler, SynchronizationContext.Current);
        }

        /// <summary>
        ///   Wraps the PropertyChangingEventHandler so that it is invoked asynchronously, and on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <param name = "syncContext">Optional. The context in which the callback is posted</param>
        /// <returns>A delegate wrapper which invokes inner delegate asynchronously</returns>
        public static PropertyChangingEventHandler AsAsync(this PropertyChangingEventHandler eventHandler,
                                                           SynchronizationContext syncContext)
        {
            if (syncContext != null)
            {
                return (o, e) =>
                           {
                               var sopcb = new SendOrPostCallback(b => eventHandler(b, e));
                               syncContext.Post(sopcb, o);
                           };
            }
            return (o, e) =>
                       {
                           var sopcb = new WaitCallback(b => eventHandler(b, e));
                           ThreadPool.QueueUserWorkItem(sopcb, o);
                       };
        }

        #endregion

        #region PropertyChangingEventHandler.AsSend

        /// <summary>
        ///   Wraps the PropertyChangingEventHandler so that it is invoked synchronously on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <returns>A delegate wrapper which invokes inner delegate synchronously</returns>
        public static PropertyChangingEventHandler AsSend(this PropertyChangingEventHandler eventHandler)
        {
            return AsSend(eventHandler, SynchronizationContext.Current);
        }

        /// <summary>
        ///   Wraps the PropertyChangingEventHandler so that it is invoked synchronously on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <param name = "syncContext">Optional. The context in which the callback is sent</param>
        /// <returns>A delegate wrapper which invokes inner delegate synchronously</returns>
        public static PropertyChangingEventHandler AsSend(this PropertyChangingEventHandler eventHandler,
                                                          SynchronizationContext syncContext)
        {
            if (syncContext != null)
            {
                return (o, e) =>
                           {
                               var sopcb = new SendOrPostCallback(b => eventHandler(b, e));
                               syncContext.Send(sopcb, o);
                           };
            }
            return eventHandler;
        }

        #endregion

        #region PropertyChangedEventHandler.AsAsync

        /// <summary>
        ///   Wraps the PropertyChangedEventHandler so that it is invoked asynchronously, and on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <returns>A delegate wrapper which invokes inner delegate asynchronously</returns>
        public static PropertyChangedEventHandler AsAsync(this PropertyChangedEventHandler eventHandler)
        {
            return AsAsync(eventHandler, SynchronizationContext.Current);
        }

        /// <summary>
        ///   Wraps the PropertyChangedEventHandler so that it is invoked asynchronously, and on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <param name = "syncContext">Optional. The context in which the callback is posted</param>
        /// <returns>A delegate wrapper which invokes inner delegate asynchronously</returns>
        public static PropertyChangedEventHandler AsAsync(this PropertyChangedEventHandler eventHandler,
                                                          SynchronizationContext syncContext)
        {
            if (syncContext != null)
            {
                return (o, e) =>
                           {
                               var sopcb = new SendOrPostCallback(b => eventHandler(b, e));
                               syncContext.Post(sopcb, o);
                           };
            }
            return (o, e) =>
                       {
                           var sopcb = new WaitCallback(b => eventHandler(b, e));
                           ThreadPool.QueueUserWorkItem(sopcb, o);
                       };
        }

        //public static  PropertyChangedEventHandler IsAsync(this PropertyChangedEventHandler eventHandler)
        //{
        //    return eventHandler.Target.GetType().    
        //}

        #endregion

        #region PropertyChangedEventHandler.AsSend

        /// <summary>
        ///   Wraps the PropertyChangedEventHandler so that it is invoked synchronously on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <returns>A delegate wrapper which invokes inner delegate synchronously</returns>
        public static PropertyChangedEventHandler AsSend(this PropertyChangedEventHandler eventHandler)
        {
            return AsSend(eventHandler, SynchronizationContext.Current);
        }

        /// <summary>
        ///   Wraps the PropertyChangedEventHandler so that it is invoked synchronously on the _callerType's thread if a SynchronizationContext is available.
        /// </summary>
        /// <param name = "eventHandler">The callback to be wrapped</param>
        /// <param name = "syncContext">Optional. The context in which the callback is sent</param>
        /// <returns>A delegate wrapper which invokes inner delegate synchronously</returns>
        public static PropertyChangedEventHandler AsSend(this PropertyChangedEventHandler eventHandler,
                                                         SynchronizationContext syncContext)
        {
            if (syncContext != null)
            {
                return (o, e) =>
                           {
                               var sopcb = new SendOrPostCallback(b => eventHandler(b, e));
                               syncContext.Send(sopcb, o);
                           };
            }
            return eventHandler;
        }

        #endregion

    }
}