﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SharpObservation.Virtual
{
    /// <summary>
    ///   Wraps a single target object, providing information about whether or not the object is present.
    /// </summary>
    /// <typeparam name = "T">The typeToMatch being wrapped</typeparam>
    /// <remarks>
    ///   DataWrapper enables target-bound virtualization scenarios, where a collection is immediately returned to a GUI for binding
    ///   event though the underlying target is not yet fully loaded. The DataWrapper acts as a placeholder for the object that is of interrest
    ///   to the GUI, and providers the GUI with notification about the loaded state of the target.
    /// </remarks>
    public class ItemReference<T> : ObservableObject, IReference<T>
        where T : class
    {
        // backing fields
        private bool _isAlive;
        private T _target;
        private readonly VirtualPage<T> _page;   

        /// <summary>
        /// Defines the observable Target property
        /// </summary>
        private static readonly Property<ItemReference<T>, T> TargetProperty = new Property<ItemReference<T>, T>(o => o.Target,
            o => o._target, (o, v) => o._target = v);

        /// <summary>
        /// Defines the observable IsAlive property
        /// </summary>
        private static readonly Property<ItemReference<T>, bool> IsAliveProperty = new Property<ItemReference<T>, bool>(o => o.IsAlive, 
            o => o._isAlive, (o,v) => o._isAlive = v);


        /// <summary>
        /// Initializes a new instance of the <see cref="ItemReference&lt;T&gt;"/> class.
        /// </summary>
        /// <param name="page">The page.</param>
        internal ItemReference(VirtualPage<T> page)
        {
            _isAlive = false;
            _target = null;
            _page = page;
        }


        /// <summary>
        ///   The current instance witin the list
        /// </summary>
        public T Target
        {
            get { return _target; }
            internal set
            {
                TargetProperty.Set(this, value);
                IsAliveProperty.Set(this, value != null);
            }
        }

        /// <summary>
        ///   Gets whether a value is assigned to Target
        /// </summary>
        public bool IsAlive
        {
            get { return _isAlive; }
        }

        public VirtualPage<T> Page
        {
            get { return _page; }
        }
    }
}
