﻿using System;
using System.Collections;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading;

namespace SharpObservation.Virtual
{
    /// <summary>
    /// Specialized list implementation that provides target virtualization. The collection is divided up into pages,
    /// and pages are dynamically GetOrAdded from the IItemsProvider when required. Stale pages are removed after a
    /// configurable period of time.
    /// Intended for use with large collections on a network or disk resource that cannot be instantiated locally
    /// due to memory consumption or GetOrAdd latency.
    /// </summary>
    /// <remarks>
    /// The IList implmentation is not fully complete, but should be sufficient for use as read only collection 
    /// target bound to a suitable ItemsControl.
    /// </remarks>
    /// <typeparam name = "T"></typeparam>
    public class VirtualCollection<T> : ObservableObject, IList<ItemReference<T>>, IList, INotifyCollectionChanged
        where T : class
    {
        #region Fields

        private object _syncRoot;
        private readonly int _pageSize = 100;
        private readonly IItemsProvider<T> _itemsProvider;
        private readonly ConcurrentDictionary<int, WeakReference<VirtualPage<T>>> _pages;
        private readonly ConcurrentQueue<Tuple<VirtualPage<T>, Timer>> _retainedPages;
        private readonly VirtualPage<T> _firstPage;
        private int _pagesBusyLoadingCount;
        private TimeSpan _retainPageTimeSpan;

        #endregion

        #region Constructor

        /// <summary>
        /// Initializes a new instance of the <see cref = "VirtualCollection{T}" /> class.
        /// </summary>
        /// <param name = "itemsProvider">The _items factory.</param>
        /// <param name = "pageSize">Size of the page.</param>
        public VirtualCollection(IItemsProvider<T> itemsProvider, int pageSize)
        {
            _itemsProvider = itemsProvider;
            _assumedCount = _pageSize = pageSize;

            var propChanged = _itemsProvider as INotifyPropertyChanged;
            if (propChanged != null)
                propChanged.PropertyChanged += new PropertyChangedEventHandler(HandleItemsProviderPropertyChanged);

            _firstPage = CreatePage(0);
            _pages = new ConcurrentDictionary<int, WeakReference<VirtualPage<T>>>();
            _retainedPages = new ConcurrentQueue<Tuple<VirtualPage<T>, Timer>>();
            _retainPageTimeSpan = new TimeSpan(0, 0, 0, 15);

            if (_itemsProvider.IsThreadSafe)
                RequestCount();
        }

        #endregion

        #region Properties & Events

        #region Property : ItemsProvider

        /// <summary>
        /// Gets the _items factory.
        /// </summary>
        /// <Value>The _items factory.</Value>
        public IItemsProvider<T> ItemsProvider
        {
            get { return _itemsProvider; }
        }

        #endregion

        #region Property : int PageSize

        /// <summary>
        /// Gets the number of items that are paged in at any given time.
        /// </summary>
        /// <Value>The size of the page.</Value>
        public int PageSize
        {
            get { return _pageSize; }
        }

        #endregion

        #region Property : int RetainPageTimeSpan

        /// <summary>
        /// Gets or Sets the minimum timespan for which loaded pages will be kept in the cache
        /// </summary>
        public TimeSpan RetainPageTimeSpan
        {
            get { return _retainPageTimeSpan; }
            set { _retainPageTimeSpan = value; }
        }

        #endregion

        #region Property : LoadingState CountState

        /// <summary>
        /// Backing Field for CountState
        /// </summary>
        private LoadingState _countState;

        /// <summary>
        /// Prototype for the LoadingState property 
        /// </summary>
        private static readonly Property<VirtualCollection<T>, LoadingState> CountStateProperty =
                  new Property<VirtualCollection<T>, LoadingState>(
                      c => c.CountState,
                      c => c._countState,
                      (c, v) => c._countState = v);

        /// <summary>
        /// Gets or Sets the LoadingState value for CountState
        /// </summary>
        public LoadingState CountState
        {
            get { return _countState; }
            private set { CountStateProperty.Set(this, value); }
        }

        #endregion

        #region Property : int AssumedCount

        private int _assumedCount;

        /// <summary>
        /// Gets or sets the assumed count.
        /// </summary>
        /// <value>The assumed count.</value>
        public int AssumedCount
        {
            get { return _assumedCount; }
            set { _assumedCount = value; }
        }

        #endregion

        #region Property : int Count

        private int _count = -1;

        private static readonly Property<VirtualCollection<T>, int> CountProperty = new Property
            <VirtualCollection<T>, int>(o => o.Count, o => o._count, (o, v) => o._count = v);

        /// <summary>
        /// Gets the number of elements contained in the list.
        /// </summary>
        /// <Value></Value>
        /// <returns>
        /// The number of elements contained in the list.
        /// </returns>
        public int Count
        {
            get
            {
                if(_countState == LoadingState.Loaded)
                    return _count;

                if (!_itemsProvider.IsThreadSafe)
                {
                    _count = ItemsProvider.Count;
                    CountState = LoadingState.Loaded;
                    return _count;
                }

                RequestCount();

                return _assumedCount;
            }
        }

        void RequestCount()
        {
            if (CountState == LoadingState.Unloaded)
            {
                CountState = LoadingState.Loading;
                ThreadPool.QueueUserWorkItem(o => RequestCountDoWork());
            }
        }

        void RequestCountDoWork()
        {
            int count = ItemsProvider.Count;
            if (count != _count)
            {
                _count = count;
                OnPropertyChanged("Count");
            }
            CountState = LoadingState.Loaded;
        }

        #endregion

        #region Event : CollectionChanged

        private static readonly EventKey<NotifyCollectionChangedEventHandler> CollectionChangedEventKey =
            new EventKey<NotifyCollectionChangedEventHandler>();


        /// <summary>
        /// Occurs when the collection changes.
        /// </summary>
        public event NotifyCollectionChangedEventHandler CollectionChanged
        {
            add { AddHandler(CollectionChangedEventKey, value); }
            remove { RemoveHandler(CollectionChangedEventKey, value); }
        }


        /// <summary>
        ///   Fires the collection reset event.
        /// </summary>
        private void FireCollectionReset()
        {
            OnCollectionChanged(new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Reset));
        }

        internal void HandleItemsNotFound(ItemReference<T>[] removed, int newCount)
        {
            if(_countState ==  LoadingState.Loaded && newCount > _count)
                CountProperty.Set(this, newCount);

            var changedArgs = new NotifyCollectionChangedEventArgs(NotifyCollectionChangedAction.Remove, removed);

            OnCollectionChanged(changedArgs);
        }

        private void OnCollectionChanged(NotifyCollectionChangedEventArgs changedArgs)
        {
            RaiseEvent(CollectionChangedEventKey, evt => evt(this, changedArgs));
            OnPropertyChanged("Item[]");
        }

        #endregion

        #region Property : this[index]

        /// <summary>
        /// Gets the item at the specified virtualIndex. This property will get
        /// the corresponding page from the IItemsProvider if required.
        /// </summary>
        /// <Value></Value>
        public ItemReference<T> this[int index]
        {
            get
            {
                if (index >= Count)
                    throw new ArgumentOutOfRangeException("index");

                int pageOffset = index % PageSize;
                var page = GetOrAddPage(index - pageOffset);
                return page[pageOffset];
            }
        }

        #endregion

        #endregion

        #region Collection

        /// <summary>
        /// Returns an enumerator that iterates through the collection.
        /// </summary>
        /// <remarks>
        /// This method should be avoided on large collections due to poor performance.
        /// </remarks>
        /// <returns>
        /// A <see cref = "IEnumerator{T}" /> that can be used to iterate through the collection.
        /// </returns>
        public IEnumerator<ItemReference<T>> GetEnumerator()
        {
            for (int i = 0; i < Count; i++)
            {
                yield return this[i];
            }
        }

        /// <summary>
        /// Returns an enumerator that iterates through a collection.
        /// </summary>
        /// <returns>
        /// An <see cref = "IEnumerator{T}" /> object that can be used to iterate through the collection.
        /// </returns>
        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }

        /// <summary>
        /// Determine whether this collection contains a given item
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns>
        /// 	<c>true</c> if [contains] [the specified item]; otherwise, <c>false</c>.
        /// </returns>
        public bool Contains(ItemReference<T> item)
        {
            return IndexOf(item) != -1;
        }

        /// <summary>
        /// Returns the indexes the of the given item
        /// </summary>
        /// <param name="item">The item.</param>
        /// <returns></returns>
        public int IndexOf(ItemReference<T> item)
        {
            var page = item.Page;
            if (page == null || page.Owner != this)
                return -1;

            var index = page.IndexOf(item);
            return index + page.PageOffset;
        }

        #endregion

        #region Page Management Methods

        void HandleItemsProviderPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == "Count")
            {
                _assumedCount = _count;
                _count = -1;
                ResetAllPages();
            }
        }

        /// <summary>
        /// Removes all loaded pages. This is useful when the count of the underlying collection changes.
        /// </summary>
        protected void ResetAllPages()
        {
            foreach (var p in _pages)
                if (p.Value.IsAlive)
                    p.Value.Target.Reset();

            FireCollectionReset();
        }

        /// <summary>
        /// Makes a request for the specified page, creating the necessary slots in the dictionary,
        /// and updating the page touch time.
        /// </summary>
        /// <param name = "pageOffset">Index of the page.</param>
        private VirtualPage<T> GetOrAddPage(int pageOffset)
        {
            if (pageOffset == 0) return _firstPage;

            var pageRef = _pages.AddOrUpdate(pageOffset, CreatePageReference,
                                             (key, val1) => val1.IsAlive ? val1 : CreatePageReference(key));

            return pageRef.Target;
        }

        /// <summary>
        /// Creates a new weak reference to a page
        /// </summary>
        /// <param name="pageOffset">The index of the page to create</param>
        /// <returns></returns>
        private WeakReference<VirtualPage<T>> CreatePageReference(int pageOffset)
        {
            var page = CreatePage(pageOffset);

            if (_retainPageTimeSpan > TimeSpan.Zero)
            {
                var timer = new Timer(VisitRetainedPages, null, _retainPageTimeSpan, TimeSpan.Zero);
                _retainedPages.Enqueue(Tuple.Create(page, timer));
            }

            return new WeakReference<VirtualPage<T>>(page);
        }

        /// <summary>
        /// Creates a new virtual page.
        /// </summary>
        /// <param name="pageOffset">Index of the page.</param>
        /// <returns></returns>
        private VirtualPage<T> CreatePage(int pageOffset)
        {
            var page = new VirtualPage<T>(pageOffset, PageSize, this);
            return page;
        }

        /// <summary>
        /// Visits the retained pages, removing strong references
        /// </summary>
        /// <param name="state">The state.</param>
        private void VisitRetainedPages(object state)
        {
            Tuple<VirtualPage<T>, Timer> item;
            _retainedPages.TryDequeue(out item);
        }

        #endregion

        #region IList Members

        /// <summary>
        /// Gets or sets the <see cref="System.Object"/> at the specified index.
        /// </summary>
        /// <value></value>
        object IList.this[int index]
        {
            get { return this[index]; }
            set { throw new NotSupportedException(); }
        }

        int IList.Add(object value)
        {
            throw new NotSupportedException();
        }

        bool IList.Contains(object value)
        {
            return Contains((ItemReference<T>)value);
        }

        public void Clear()
        {
            throw new NotImplementedException();
        }

        int IList.IndexOf(object value)
        {
            return IndexOf((ItemReference<T>)value);
        }


        /// <summary>
        /// Gets an object that can be used to synchronize access to the <see cref="T:System.Collections.ICollection"/>.
        /// </summary>
        /// <value></value>
        /// <returns>An object that can be used to synchronize access to the <see cref="T:System.Collections.ICollection"/>.</returns>
        public object SyncRoot
        {
            get
            {
                if (this._syncRoot == null)
                {
                    Interlocked.CompareExchange<object>(ref this._syncRoot, new object(), null);
                }
                return this._syncRoot;
            }
        }

        /// <summary>
        /// Gets a value indicating whether access to the <see cref="T:System.Collections.ICollection"/> is synchronized (thread safe).
        /// </summary>
        /// <value></value>
        /// <returns>true if access to the <see cref="T:System.Collections.ICollection"/> is synchronized (thread safe); otherwise, false.</returns>
        public bool IsSynchronized
        {
            get { return false; }
        }

        /// <summary>
        /// Gets a Value indicating whether the <see cref = "System.Collections.IList" /> has a fixed size.
        /// </summary>
        /// <Value></Value>
        /// <returns>Always false.</returns>
        public bool IsFixedSize
        {
            get { return false; }
        }

        /// <summary>
        /// Gets a value indicating whether this instance is read only.
        /// </summary>
        /// <value>
        /// <c>true</c> if this instance is read only; otherwise, <c>false</c>.
        /// </value>
        public bool IsReadOnly
        {
            get { return true; }
        }

        #endregion

        #region IList<ItemReference<T>> Members

        void IList.Insert(int index, object value)
        {
            throw new NotSupportedException();
        }

        void IList.Remove(object value)
        {
            throw new NotSupportedException();
        }

        void IList<ItemReference<T>>.Insert(int index, ItemReference<T> item)
        {
            throw new NotSupportedException();
        }

        void IList<ItemReference<T>>.RemoveAt(int index)
        {
            throw new NotSupportedException();
        }

        ItemReference<T> IList<ItemReference<T>>.this[int index]
        {
            get { return this[index]; }
            set { throw new NotSupportedException(); }
        }

        void ICollection<ItemReference<T>>.Add(ItemReference<T> item)
        {
            throw new NotSupportedException();
        }

        void ICollection<ItemReference<T>>.Clear()
        {
            throw new NotSupportedException();
        }

        void ICollection<ItemReference<T>>.CopyTo(ItemReference<T>[] array, int arrayIndex)
        {
            throw new NotSupportedException();
        }

        bool ICollection<ItemReference<T>>.Remove(ItemReference<T> item)
        {
            throw new NotSupportedException();
        }

        void IList.RemoveAt(int index)
        {
            throw new NotSupportedException();
        }

        void ICollection.CopyTo(Array array, int index)
        {
            throw new NotSupportedException();
        }
         
        #endregion

    }
}
