﻿#region

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Threading;

#endregion

namespace SharpObservation.Virtual
{
    /// <summary>
    ///   Represents a page within a collection of virtualizable data
    /// </summary>
    /// <typeparam name = "T">The type of item being virtualized</typeparam>
    public class VirtualPage<T> : ObservableObject, IEnumerable<ItemReference<T>>
        where T : class
    {
        private readonly ItemReference<T>[] _items;
        private readonly VirtualCollection<T> _owner;
        private readonly int _pageOffset;

        /// <summary>
        ///   Initializes a new instance of the <see cref = "VirtualPage{T}" /> class.
        /// </summary>
        /// <param name = "pageOffset">The first index.</param>
        /// <param name = "pageLength">Length of the page.</param>
        /// <param name = "owner">The collection owning this page.</param>
        public VirtualPage(int pageOffset, int pageLength, VirtualCollection<T> owner)
        {
            _pageOffset = pageOffset;
            _count = pageLength;
            _items = Enumerable.Range(0, pageLength).Select(i => CreateReference()).ToArray();
            _owner = owner;
        }

        #region Owner Property

        internal VirtualCollection<T> Owner
        {
            get { return _owner; }
        }

        #endregion

        #region IsInUse Property

        /// <summary>
        ///   Prototype for the bool property
        /// </summary>
        private static readonly Property<VirtualPage<T>, bool> IsInUseProperty =
            new Property<VirtualPage<T>, bool>(
                c => c.IsInUse,
                c => c._inUseCount > 0,
                (c, v) => { });

        private int _inUseCount;

        /// <summary>
        ///   Gets or Sets the bool value for IsInUse
        /// </summary>
        public bool IsInUse
        {
            get { return _inUseCount > 0; }
        }


        private void WrapperActivationHandler(object sender, IsNotificationActiveChangedEventArgs e)
        {
            var oldCount = _inUseCount;

            _inUseCount += e.IsNotificationActive ? 1 : -1;

            if (_inUseCount == 0 || oldCount == 0)
                IsInUseProperty.Set(this, _inUseCount != 0);
        }

        #endregion

        #region State Property

        /// <summary>
        ///   Prototype for the LoadingState property
        /// </summary>
        private static readonly Property<VirtualPage<T>, LoadingState> StateProperty =
            new Property<VirtualPage<T>, LoadingState>(
                c => c.State,
                c => c._state,
                (c, v) => c._state = v);

        /// <summary>
        ///   Backing Field for State
        /// </summary>
        private LoadingState _state;

        /// <summary>
        ///   Gets or Sets the LoadingState value for State
        /// </summary>
        public LoadingState State
        {
            get { return _state; }
            set { StateProperty.Set(this, value); }
        }

        #endregion

        #region Count Property

        /// <summary>
        ///   Prototype for the int property
        /// </summary>
        private static readonly Property<VirtualPage<T>, int> CountProperty =
            new Property<VirtualPage<T>, int>(
                c => c.Count,
                c => c._count,
                (c, v) => c._count = v);

        /// <summary>
        ///   Backing Field for PageLength
        /// </summary>
        private int _count;

        /// <summary>
        ///   Gets the value for PageLength
        /// </summary>
        public int Count
        {
            get { return _count; }
            private set { CountProperty.Set(this, value); }
        }

        #endregion

        /// <summary>
        ///   Gets the index of the first item in the page
        /// </summary>
        /// <value>The first index.</value>
        public int PageOffset
        {
            get { return _pageOffset; }
        }

        /// <summary>
        ///   Gets the <see cref = "SharpObservation.Virtual.ItemReference&lt;T&gt;" /> at the specified index.
        /// </summary>
        /// <value></value>
        public ItemReference<T> this[int index]
        {
            get
            {
                if (index >= _count)
                    throw new ArgumentOutOfRangeException("index");

                if (_state == LoadingState.Unloaded)
                    LoadData();

                return _items[index];
            }
        }

        #region IEnumerable<ItemReference<T>> Members

        public IEnumerator<ItemReference<T>> GetEnumerator()
        {
            if (_state == LoadingState.Unloaded)
                LoadDataWork();

            return _items.Take(_count).GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return _items.GetEnumerator();
        }

        #endregion

        private void LoadData()
        {
            if (_state == LoadingState.Unloaded)
            {
                StateProperty.Set(this, LoadingState.Loading);

                if (_owner.ItemsProvider.IsThreadSafe)
                    ThreadPool.QueueUserWorkItem(o => LoadDataWork(), null);
                else
                    LoadDataWork();
            }
        }

        private void LoadDataWork()
        {
            // Consume all items into a local array before pinging back to the UI thread...
            int i = 0;
            foreach (var item in _owner.ItemsProvider.GetRange(_pageOffset, _count))
                _items[i++].Target = item;

            if(i < _count)
            {
                var removed = new ItemReference<T>[_count - i];
                for (int j = i; j < _count; j++)
                    removed[j-i] = _items[i];

                _owner.HandleItemsNotFound(removed, _pageOffset + i);

            }

            CountProperty.Set(this, i);
            StateProperty.Set(this, LoadingState.Loaded);
        }

        internal void Reset()
        {
            StateProperty.Set(this, LoadingState.Unloaded);
            foreach (var item in _items)
                item.Target = null;
        }

        /// <summary>
        ///   Creates a new DataWrapper
        /// </summary>
        private ItemReference<T> CreateReference()
        {
            var wrapper = new ItemReference<T>(this);
            wrapper.IsNotificationActiveChanged += WrapperActivationHandler;
            return wrapper;
        }

        public int IndexOf(ItemReference<T> value)
        {
            return Array.IndexOf(_items, value);
        }
    }
}