﻿#region

using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Reflection.Emit;

#endregion

namespace SharpObservation
{
    /// <summary>
    /// Extends methods to Delegate for creating Weak Reference delegates.
    /// </summary>
    /// <remarks>
    /// <para>
    /// .NET uses garbage collection to reduce the likelyhood memory leaks by automatically detecting abandoned objects and 
    /// releasing the memory used by them. The use of events and delegates has a tendency to defeat this objective, because
    /// it is typically presentation code which attaches delegates to events, and in scenarios where the objects with the events being attached to 
    /// out live the user interface objects which "listen" to them, a memory leak occurs. This is because the delegates attached to the 
    /// long-lived object's events hold references to the listening objects, preventing them from being garbage collected.
    /// </para><para>
    /// There are only two solutions to this problem. Either:
    /// <list type="bullet">
    /// <item>Hand-code the tracking of all delegates, and dettach them all upon closing a form/window/control, or</item>
    /// <item>Implement the delegates in a manner that does not keep the listener alive - i.e. create a weak delegate</item>
    /// </list></para>
    /// <para>The WeakDelegate type exposes functionality to achieve exactly that... it converts an existing delegate into a weak version
    /// of itself. Moreover, the delegate also retains a callback function which allows it to dettach itself from it's source event
    /// if it detects that the target (the listener) has been garbage collected or disposed.
    /// </para>
    /// <para>
    /// There are three possible states in which a delegate can find itself, and the WeakDelegate implementaion handles all three:
    /// <list type="number">
    /// <item>The source and the listener are both alive. When the source evnet fires, the listener's method is invoked.</item>
    /// <item>The source is garbage collected. The delegate automatically becomes elligible for collection.</item>
    /// <item>The listener is garbage collected. The next time the listener raises its event, the delegate detaches itself.</item>
    /// </list>
    /// </para>
    /// <para>The follow example shows a long lived singleton, which exposes an observable property "Name". The UI form listens to the
    /// Name property and reacts to any changes. Notice how MakeWeak is used to prevent the form from being kept alive by the singleton:
    /// <code>
    /// public class SingletonExample : ObservableObject 
    /// {
    ///     private static SingletonExample _this;
    /// 
    ///     public static SingletonExample Current { get { return _this ? (_this = new SingletonExample() ); } }
    /// 
    ///        private string _name;
    ///
    ///        private static readonly Property&lt;NotifyDummy, string&gt; NameProperty = new Property&lt;NotifyDummy, string&gt;(
    ///                      c => c.Name,               
    ///                      c => c._name,              
    ///                      (c, v) => c._name = v);    
    ///
    ///        public string Name
    ///        {
    ///            get { return _name; }
    ///            set { NameProperty.Set(this, value); }       
    ///        }
    /// }
    /// 
    /// 
    /// public class MyForm : Form
    /// {
    ///     public void Initialize()
    ///     {
    ///         var watchedObject =  SingletonExample.Current;
    ///         var myDelegate = new PropertyChangedEventHandler( SingletonNameChanged );
    /// 
    ///         watchedObject.PropertyChanged += myDelegate.MakeWeak();
    ///     }
    /// 
    ///     public void SingletonNameChanged(object sender, PropertyChangedEventArgs e)
    ///     {
    ///         if(e.PropertyName == "Name")
    ///             this.Title = ((SingletonExample)sender).Name;
    ///     } 
    /// }
    /// </code>
    /// </para>
    /// </remarks>
    public static class WeakDelegate
    {
        #region Statics

        private static readonly AssemblyBuilder DynamicAssembly;
        private static readonly ModuleBuilder DynamicModule;

        /// <summary>
        ///   Static reference, used by MethodBoundTemplate.DefineInvoke
        /// </summary>
        private static readonly ConstructorInfo WeakReferenceConstructor = typeof(WeakReference).GetConstructor(new[] { typeof(object) });

        /// <summary>
        ///   Static reference, used by MethodBoundTemplate.DefineInvoke
        /// </summary>
        private static readonly MethodInfo WeakReferenceTargetGetter =
            typeof(WeakReference).GetProperty("Target").GetGetMethod();

        /// <summary>
        ///   Static reference, used by MethodBoundTemplate.DefineInvoke
        /// </summary>
        private static readonly MethodInfo DisposalInfoIsDisposedGetter =
            typeof(IDisposalInfo).GetProperty("IsDisposed").GetGetMethod();


        static WeakDelegate()
        {
            DynamicAssembly =
                AppDomain.CurrentDomain.DefineDynamicAssembly(new AssemblyName("SharpObservations.Weakness"),
                                                              AssemblyBuilderAccess.Run);
            DynamicModule = DynamicAssembly.DefineDynamicModule("WeakDelegates");
        }

        #endregion

        #region MakeWeak extension methods and overloads

        /// <summary>
        /// Returns a weak-reference version of a delegate
        /// </summary>
        /// <param name = "aDelegate">The delegate to convert to weak referencing</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static TDelegate MakeWeak<TDelegate>(this Delegate aDelegate)
            where TDelegate : class
        {
            return MethodBoundTemplate<TDelegate>.CreateWeak(aDelegate as TDelegate, null);
        }

        /// <summary>
        ///   Returns a weak-reference version of a delegate
        /// </summary>
        /// <param name = "aDelegate">The delegate to convert to weak referencing</param>
        /// <param name = "fnUnregister">The unregister action to invoke if the target is garbage collected</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static TDelegate MakeWeak<TDelegate>(this Delegate aDelegate, Action<TDelegate> fnUnregister)
            where TDelegate : class
        {
            return MethodBoundTemplate<TDelegate>.CreateWeak(aDelegate as TDelegate, fnUnregister);
        }

        /// <summary>
        /// Returns a weak-reference version of an event handler
        /// </summary>
        /// <param name = "aDelegate">The delegate to convert to weak referencing</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static EventHandler<T> MakeWeak<T>(this EventHandler<T> aDelegate)
            where T : EventArgs
        {
            return MethodBoundTemplate<EventHandler<T>>.CreateWeak(aDelegate, null);
        }

        /// <summary>
        ///   Returns a weak-reference version of a event handler
        /// </summary>
        /// <param name = "handler">The delegate to convert to weak referencing</param>
        /// <param name = "fnUnregister">The unregister action to invoke if the target is garbage collected</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static EventHandler<T> MakeWeak<T>(this EventHandler<T> handler, Action<EventHandler<T>> fnUnregister)
            where T : EventArgs
        {
            return MethodBoundTemplate<EventHandler<T>>.CreateWeak(handler, fnUnregister);
        }


        /// <summary>
        ///   Returns a weak-reference version of a PropertyChangedEventHandler
        /// </summary>
        /// <param name = "handler">The delegate to convert to weak referencing</param>
        /// <param name = "fnUnregister">The unregister action to invoke if the target is garbage collected</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static PropertyChangedEventHandler MakeWeak(this PropertyChangedEventHandler handler, Action<PropertyChangedEventHandler> fnUnregister)
        {
            return MethodBoundTemplate<PropertyChangedEventHandler>.CreateWeak(handler, fnUnregister);
        }

        /// <summary>
        /// Returns a weak-reference version of a PropertyChangedEventHandler
        /// </summary>
        /// <param name = "handler">The delegate to convert to weak referencing</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static PropertyChangedEventHandler MakeWeak(this PropertyChangedEventHandler handler)
        {
            return MethodBoundTemplate<PropertyChangedEventHandler>.CreateWeak(handler, null);
        }


        /// <summary>
        ///   Returns a weak-reference version of a PropertyChangingEventHandler
        /// </summary>
        /// <param name = "handler">The delegate to convert to weak referencing</param>
        /// <param name = "fnUnregister">The unregister action to invoke if the target is garbage collected</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static PropertyChangingEventHandler MakeWeak(this PropertyChangingEventHandler handler, Action<PropertyChangingEventHandler> fnUnregister)
        {
            return MethodBoundTemplate<PropertyChangingEventHandler>.CreateWeak(handler, fnUnregister);
        }

        /// <summary>
        /// Returns a weak-reference version of a PropertyChangingEventHandler
        /// </summary>
        /// <param name = "handler">The delegate to convert to weak referencing</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static PropertyChangingEventHandler MakeWeak(this PropertyChangingEventHandler handler)
        {
            return MethodBoundTemplate<PropertyChangingEventHandler>.CreateWeak(handler, null);
        }


        /// <summary>
        ///   Returns a weak-reference version of a delegate
        /// </summary>
        /// <param name = "aDelegate">The delegate to convert to weak referencing</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static TDelegate MakeWeak<TDelegate>(TDelegate aDelegate)
            where TDelegate : class
        {
            return MethodBoundTemplate<TDelegate>.CreateWeak(aDelegate, null);
        }

        /// <summary>
        ///   Returns a weak-reference version of a delegate
        /// </summary>
        /// <param name = "aDelegate">The delegate to convert to weak referencing</param>
        /// <param name = "fnUnregister">The unregister action to invoke if the target is garbage collected</param>
        /// <returns>A new, weak referencing delegate</returns>
        public static TDelegate MakeWeak<TDelegate>(TDelegate aDelegate, Action<TDelegate> fnUnregister)
            where TDelegate : class
        {
            return MethodBoundTemplate<TDelegate>.CreateWeak(aDelegate, fnUnregister);
        }

        #endregion

        #region IsWeak

        /// <summary>
        ///   Gets whether the delegate is weak referenced
        /// </summary>
        /// <param name = "anyDelegate"></param>
        /// <returns>True if the delegate is a weak reference</returns>
        public static bool IsWeak(this Delegate anyDelegate)
        {
            return (anyDelegate.Target as WeakReference) != null;

            //var target = anyDelegate.Target;
            //if (target == null)
            //    return false;

            //return target.GetType().Assembly.FullName == DynamicAssembly.FullName;
        }

        #endregion

        #region Nested type: DelegateConverter

        /// <summary>
        ///   Defines the delegate for converting a delegate to it's weak refenced counterpart
        /// </summary>
        /// <param name = "original">The original delegate, to be converted</param>
        /// <param name = "fnUnregister">A callback which will be invoked if the target is garbage collected</param>
        /// <returns>A weak refencing delegate which does not keep it's target alive.</returns>
        [System.Runtime.CompilerServices.CompilerGenerated]
        private delegate TDelegate DelegateConverter<TDelegate>(TDelegate original, Action<TDelegate> fnUnregister);

        #endregion

        #region Nested type: MethodBoundTemplate

        /// <summary>
        ///   Encapsulates the logic for a converter which converts a MethodInfo bound delegate into a weak referencing version of itself
        /// </summary>
        /// <typeparam name = "TDelegate">The type of delegate to be converted</typeparam>
        [System.Runtime.CompilerServices.CompilerGenerated]
        private class MethodBoundTemplate<TDelegate>
            where TDelegate : class
        {
            #region Nested Types

            #region Nested type: InnerConverter

            /// <summary>
            ///   Defines the signature for the IL-generated delegate converter method
            /// </summary>
            /// <param name = "original">The original delegate</param>
            /// <param name = "fnUnregister">The unregister callback</param>
            /// <param name = "openDlegate">The type to be used as an open version of TDelegate</param>
            /// <returns>A weak referencing delegate</returns>
            [System.Runtime.CompilerServices.CompilerGenerated]
            private delegate TDelegate InnerConverter(
                TDelegate original, Action<TDelegate> fnUnregister, Delegate openDlegate);

            #endregion

            #region Nested type: WeakDelegateBuilder

            /// <summary>
            ///   Encapsulates construction of new type to implement weak referencing
            /// </summary>
            /// <remarks>
            ///   The type produced by this builder is equivilant to the following:
            /// 
            ///   public class WeakDelegate : WeakReference
            ///   {
            ///      OpenDelegate _openDelegate;
            ///      Action[Delegate] _unregister;
            /// 
            ///      public WeakDelegate(TDelegate aDelegate, Action[TDelegate] fnUnregister) : base(aDelegate.Target)
            ///      {
            ///         _openDelegate = aDelegate;
            ///         _unregister = fnUnregister;
            ///      }
            /// 
            ///      public void Invoke (Arg1 a, Arg2 b, ...)    --- signature matches TDelegate's signature
            ///      {
            ///          var target = this._Target;
            ///          if(target != null)
            ///          {
            ///             var asDisposalInfo = target as DisposalInfo;
            ///             if(asDisposalInfo != null &amp;&amp; !asDisposalInfo.IsDisposed)
            ///             {
            ///                 _openDelegate(target, a, b, ...);
            ///                 return;
            ///             }
            ///          }
            ///          if(_unregister != null)
            ///          {
            ///             _unregister(new TDelegate(this.Invoke));
            ///             _unregister = null;
            ///          }
            ///      }
            ///   }
            /// </remarks>
            [System.Runtime.CompilerServices.CompilerGenerated]
            private class WeakDelegateBuilder
            {
                private readonly TypeBuilder _classBuilder;
                private readonly FieldBuilder _delegateField;
                private readonly MethodInfo _method;
                private readonly FieldBuilder _unregisterField;
                private readonly Type _weakType;

                public InnerConverter Inner;
                public Delegate OpenDelegate;
                private Type _openDelegateType;

                public WeakDelegateBuilder(MethodInfo method)
                {
                    _method = method;

                    _classBuilder =
                        DynamicModule.DefineType(string.Format("{0}.{1}", method.DeclaringType.Name, method.Name),
                                                 TypeAttributes.Class | TypeAttributes.Public,
                                                 typeof(WeakReference));

                    _unregisterField = _classBuilder.DefineField("_unregister", typeof(Action<TDelegate>),
                                                                 FieldAttributes.Public);

                    InitializeOpenDelegate();

                    _delegateField = _classBuilder.DefineField("_delegate", _openDelegateType, FieldAttributes.Public);

                    DefineConstructor();
                    DefineInvoke();

                    _weakType = _classBuilder.CreateType();

                    DefineConverter();
                }

                /// <summary>
                ///   Produces a new delegate type which is signature compatible with an open TDelegate
                /// </summary>
                private void InitializeOpenDelegate()
                {
                    var genericAction = Type.GetType(string.Format("System.Action`{0}", 1 + DelgParamTypes.Length));
                    var argTypes = new Type[DelgParamTypes.Length + 1];
                    argTypes[0] = _method.DeclaringType;
                    DelgParamTypes.CopyTo(argTypes, 1);
                    _openDelegateType = genericAction.MakeGenericType(argTypes);
                    OpenDelegate = Delegate.CreateDelegate(_openDelegateType, null, _method, true);
                }


                /// <summary>
                ///   Emits the constructor for the new type.
                /// </summary>
                private void DefineConstructor()
                {
                    var weakTypeCtorArgs = new[] { typeof(TDelegate), typeof(Action<TDelegate>), _openDelegateType };
                    var weakTypeCtor = _classBuilder.DefineConstructor(MethodAttributes.Public,
                                                                       CallingConventions.Standard,
                                                                       weakTypeCtorArgs);
                    var il = weakTypeCtor.GetILGenerator();
                    var ilTemp = il.DeclareLocal(typeof(object));

                    // construct (TDelegate delegate, Action<TDelegate> unregister, OpenDelegateType aDelegate) : base (delegate.Target)
                    // {                  
                    il.Emit(OpCodes.Ldarg_1); // delegate
                    il.Emit(OpCodes.Call, typeof(TDelegate).GetProperty("Target").GetGetMethod()); // .Target
                    il.Emit(OpCodes.Stloc, ilTemp);

                    il.Emit(OpCodes.Ldarg_0); // must pass this to base()
                    il.Emit(OpCodes.Ldloc, ilTemp);
                    il.Emit(OpCodes.Call, WeakReferenceConstructor);

                    //  this._Unregister = unregister
                    il.Emit(OpCodes.Ldarg_0);
                    il.Emit(OpCodes.Ldarg_2);
                    il.Emit(OpCodes.Stfld, _unregisterField);

                    //  this._delegate = aDelegate
                    il.Emit(OpCodes.Ldarg_0);
                    il.Emit(OpCodes.Ldarg_3);
                    il.Emit(OpCodes.Stfld, _delegateField);
                    // }
                    il.Emit(OpCodes.Ret);

                }

                /// <summary>
                ///   Emits the Invoke method for the new type
                /// </summary>
                private void DefineInvoke()
                {
                    #region Preamble

                    var invokeMethod = _classBuilder.DefineMethod("Invoke", MethodAttributes.Public,
                                                                  DelgSignature.ReturnType, DelgParamTypes);
                    var il = invokeMethod.GetILGenerator();
                    var ilTarget = il.DeclareLocal(typeof(object));
                    var ilAsDisposalInfo = il.DeclareLocal(typeof(IDisposalInfo));
                    var ilUnregister = il.DeclareLocal(typeof(Action<TDelegate>));
                    var ilEndLabel = il.DefineLabel();
                    var ilIsNullLabel = il.DefineLabel();
                    var ilTargetOkLabel = il.DefineLabel();

                    #endregion


                    // var target = this.Target
                    il.Emit(OpCodes.Ldarg_0);
                    il.Emit(OpCodes.Callvirt, WeakReferenceTargetGetter); // this.Target
                    il.Emit(OpCodes.Stloc, ilTarget); // store in local variable

                    // if(target != null)
                    // {
                    il.Emit(OpCodes.Ldloc, ilTarget);
                    il.Emit(OpCodes.Brfalse, ilIsNullLabel);

                    // var disp = target as IDisposalInfo
                    il.Emit(OpCodes.Ldloc, ilTarget);
                    il.Emit(OpCodes.Isinst, typeof(IDisposalInfo));
                    il.Emit(OpCodes.Stloc, ilAsDisposalInfo);

                    // if(disp == null ) goto: TargetOK
                    il.Emit(OpCodes.Ldloc, ilAsDisposalInfo);
                    il.Emit(OpCodes.Brfalse_S, ilTargetOkLabel);

                    // if(disp.IsDisposed) goto: IsNullLabel
                    il.Emit(OpCodes.Ldloc, ilAsDisposalInfo);
                    il.Emit(OpCodes.Callvirt, DisposalInfoIsDisposedGetter);
                    il.Emit(OpCodes.Brtrue_S, ilIsNullLabel);

                    // TargetOK:
                    il.MarkLabel(ilTargetOkLabel);

                    //      _delegate( @target, parm1, parm2 ...);
                    il.Emit(OpCodes.Ldarg_0);
                    il.Emit(OpCodes.Ldfld, _delegateField);

                    il.Emit(OpCodes.Ldloc, ilTarget);

                    for (short i = 0; i < DelgParams.Length; i++)
                        il.Emit(OpCodes.Ldarg, i + 1);

                    il.Emit(OpCodes.Tailcall);
                    il.Emit(OpCodes.Callvirt, _openDelegateType.GetMethod("Invoke"));
                    il.Emit(OpCodes.Ret);

                    // } else
                    // ilIsNullLabel:
                    il.MarkLabel(ilIsNullLabel);

                    // var unregister = this._unregister;
                    il.Emit(OpCodes.Ldarg_0); // this
                    il.Emit(OpCodes.Ldfld, _unregisterField); // ._unregister
                    il.Emit(OpCodes.Stloc, ilUnregister);

                    // if(unregister != null)
                    // {
                    il.Emit(OpCodes.Ldloc, ilUnregister);
                    il.Emit(OpCodes.Brfalse_S, ilEndLabel); // == null ? goto end

                    //      unregister(this, invoker);
                    il.Emit(OpCodes.Ldarg_0);
                    il.Emit(OpCodes.Ldfld, _unregisterField);

                    // new TDelegate(this.Invoke)
                    il.Emit(OpCodes.Ldarg_0);
                    il.Emit(OpCodes.Ldftn, invokeMethod);
                    il.Emit(OpCodes.Newobj, DelegateConstructor);

                    il.Emit(OpCodes.Callvirt, typeof(Action<TDelegate>).GetMethod("Invoke")); // call unregister

                    //      this._unregister == null;
                    il.Emit(OpCodes.Ldarg_0);
                    il.Emit(OpCodes.Ldnull);
                    il.Emit(OpCodes.Stfld, _unregisterField); // nullify unregister

                    // }
                    il.MarkLabel(ilEndLabel);

                    il.Emit(OpCodes.Ret);
                }

                /// <summary>
                ///   Produces a delegate which constructs a new weak referencing type
                /// </summary>
                private void DefineConverter()
                {
                    var args = new[] { typeof(TDelegate), typeof(Action<TDelegate>), typeof(Delegate) };
                    var dynamicMethod = new DynamicMethod("Construct", typeof(TDelegate), args);

                    var weakTypeCtorArgs = new[] { typeof(TDelegate), typeof(Action<TDelegate>), _openDelegateType };
                    var weakTypeCtor = _weakType.GetConstructor(weakTypeCtorArgs);

                    var il = dynamicMethod.GetILGenerator();

                    //  TDelegate Construct(TDelegate delegate, Action<TDelegate> unregister, _openDelegate)
                    //  {
                    //      ilThis = new WeakDelegate(delegate, unregister, _openDelegate) 
                    il.Emit(OpCodes.Ldarg_0); // push delegate
                    il.Emit(OpCodes.Ldarg_1); // push unregister
                    il.Emit(OpCodes.Ldarg_2); // push _openDelegate
                    il.Emit(OpCodes.Castclass, _openDelegateType); // .. as OpenDelegateType
                    il.Emit(OpCodes.Newobj, weakTypeCtor); // new object on stack

                    //      return new TDelegate( ilThis, Invoke );
                    //  }
                    il.Emit(OpCodes.Ldftn, _weakType.GetMethod("Invoke")); // add invoke arg
                    il.Emit(OpCodes.Newobj, typeof(TDelegate).GetConstructor(new[] { typeof(object), typeof(IntPtr) }));
                    il.Emit(OpCodes.Ret);

                    Inner = (InnerConverter)dynamicMethod.CreateDelegate(typeof(InnerConverter));
                }
            }

            #endregion

            #endregion


            #region Fields

            /// <summary>
            ///   Dictionary of all method-bound templates for this delegate.
            /// </summary>
            private static readonly ConcurrentDictionary<MethodInfo, DelegateConverter<TDelegate>> Templates =
                new ConcurrentDictionary<MethodInfo, DelegateConverter<TDelegate>>();

            private static readonly MethodInfo DelgSignature = typeof(TDelegate).GetMethod("Invoke");
            private static readonly ParameterInfo[] DelgParams = DelgSignature.GetParameters();
            private static readonly Type[] DelgParamTypes = DelgParams.Select(p => p.ParameterType).ToArray();

            private static readonly ConstructorInfo DelegateConstructor =
                typeof(TDelegate).GetConstructor(new[] { typeof(object), typeof(IntPtr) });

            private static bool _isDelegateVerified;

            private readonly InnerConverter _innerConverter;
            private readonly Delegate _openDelegate;

            #endregion

            #region Methods


            /// <summary>
            ///   Returns a weak-reference version of a delegate
            /// </summary>
            /// <param name = "aDelegate">The delegate to convert to weak referencing</param>
            /// <param name = "fnUnregister">The unregister action to invoke if the target is garbage collected</param>
            /// <returns>A new, weak referencing delegate</returns>
            public static TDelegate CreateWeak(TDelegate aDelegate, Action<TDelegate> fnUnregister)
            {
                var asDelegate = aDelegate as MulticastDelegate;

                if (asDelegate == null)
                    throw new GenericArgumentException(@"TDelegate must be a delegate type");

                var target = asDelegate.Target;

                // if delegate is Static return delegate. Is delegate already weak? return delegate
                if (target == null || target is WeakReference)
                    return aDelegate;

                if (!_isDelegateVerified)
                {
                    if (DelgParams.Any(p => p.IsOptional || p.IsRetval))
                        throw new GenericArgumentException(@"Delegates with ref (ByRef) out out arguments are not supported by the weak reference pattern.");

                    if (DelgSignature.ReturnType != typeof(void))
                        throw new GenericArgumentException(@"Delegates with return values are not supported by the weak reference pattern.");

                    _isDelegateVerified = true;
                }

                var converter = Templates.GetOrAdd(asDelegate.Method, m => (new MethodBoundTemplate<TDelegate>(m)).ConvertDelegateToWeak);

                return converter(aDelegate, fnUnregister);
            }

            /// <summary>
            ///   Constructs a new instance, bound to specific MethodInfo
            /// </summary>
            /// <param name = "method">The delegate's underlying method</param>
            private MethodBoundTemplate(MethodInfo method)
            {
                var args = new WeakDelegateBuilder(method);
                _innerConverter = args.Inner;
                _openDelegate = args.OpenDelegate;
            }

            /// <summary>
            ///   Returns a weak-reference version of a delegate
            /// </summary>
            /// <param name = "original">The delegate to convert to weak referencing</param>
            /// <param name = "fnUnregister">The unregister action to invoke if the target is garbage collected</param>
            /// <returns>A weak referencing delegate</returns>
            private TDelegate ConvertDelegateToWeak(TDelegate original, Action<TDelegate> fnUnregister)
            {
                return _innerConverter(original, fnUnregister, _openDelegate);
            }

            #endregion
        }

        #endregion

    }
}